@@ -1,6 +1,8 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002-2020  Tim Waugh <twaugh@redhat.com>
+ * Authors: Tim Waugh <twaugh@redhat.com>
+ *          Zdenek Dohnal <zdohnal@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
  */
 
 #include "cupsppd.h"
@@ -209,18 +211,25 @@ PPD_new (PyTypeObject *type, PyObject *args, PyObject *kwds)
 static int
 PPD_init (PPD *self, PyObject *args, PyObject *kwds)
 {
-  const char *filename;
+  PyObject *filenameobj;
+  char *filename;
 
-  if (!PyArg_ParseTuple (args, "s", &filename))
+  if (!PyArg_ParseTuple (args, "O", &filenameobj))
+    return -1;
+
+  if (UTF8_from_PyObj (&filename, filenameobj) == NULL)
     return -1;
 
   self->file = fopen (filename, "r");
   if (!self->file) {
     PyErr_SetString (PyExc_RuntimeError, "fopen failed");
+    free (filename);
     return -1;
   }
 
+  debugprintf ("+ PPD %p %s (fd %d)\n", self, filename, fileno (self->file));
   self->ppd = ppdOpenFile (filename);
+  free (filename);
   if (!self->ppd) {
     fclose (self->file);
     self->file = NULL;
@@ -235,8 +244,12 @@ PPD_init (PPD *self, PyObject *args, PyObject *kwds)
 static void
 PPD_dealloc (PPD *self)
 {
-  if (self->file)
+  if (self->file) {
+    debugprintf ("- PPD %p (fd %d)\n", self, fileno (self->file));
     fclose (self->file);
+  } else
+    debugprintf ("- PPD %p (no fd)\n", self);
+
   if (self->ppd)
     ppdClose (self->ppd);
   if (self->conv_from)
@@ -244,7 +257,7 @@ PPD_dealloc (PPD *self)
   if (self->conv_to)
     iconv_close (*self->conv_to);
 
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 /////////
@@ -255,16 +268,82 @@ static PyObject *
 PPD_localize (PPD *self)
 {
   if (!ppdLocalize (self->ppd))
-    return Py_None;
+    Py_RETURN_NONE;
   return PyErr_SetFromErrno (PyExc_RuntimeError);
 }
 
+static PyObject *
+PPD_localizeIPPReason (PPD *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *ret;
+  PyObject *reasonobj;
+  PyObject *schemeobj = NULL;
+  char *reason;
+  char *scheme = NULL;
+  char *buffer;
+  const size_t bufsize = 1024;
+  static char *kwlist[] = { "reason", "scheme", NULL };
+
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &reasonobj, &schemeobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&reason, reasonobj) == NULL)
+    return NULL;
+
+  if (schemeobj)
+    if (UTF8_from_PyObj (&scheme, schemeobj) == NULL) {
+      free (reason);
+      return NULL;
+    }
+
+  buffer = malloc (bufsize);
+  if (ppdLocalizeIPPReason (self->ppd, reason, scheme, buffer, bufsize))
+  {
+    ret = make_PyUnicode_from_ppd_string (self, buffer);
+  } else {
+    Py_RETURN_NONE;
+  }
+
+  free (reason);
+  if (scheme)
+    free (scheme);
+  free (buffer);
+  return ret;
+}
+
+static PyObject *
+PPD_localizeMarkerName (PPD *self, PyObject *args)
+{
+  PyObject *ret;
+  PyObject *nameobj;
+  char *name;
+  const char *lname;
+
+  if (!PyArg_ParseTuple (args, "O", &nameobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
+    return NULL;
+
+  lname = ppdLocalizeMarkerName (self->ppd, name);
+  free (name);
+
+  if (lname != NULL)
+  {
+    ret = make_PyUnicode_from_ppd_string (self, lname);
+  } else {
+    Py_RETURN_NONE;
+  }
+
+  return ret;
+}
+
 static PyObject *
 PPD_markDefaults (PPD *self)
 {
   ppdMarkDefaults (self->ppd);
-  Py_INCREF (Py_None);
-  return Py_None;
+  Py_RETURN_NONE;
 }
 
 static PyObject *
@@ -299,20 +378,25 @@ PPD_markOption (PPD *self, PyObject *args)
 static PyObject *
 PPD_conflicts (PPD *self)
 {
-  return PyInt_FromLong (ppdConflicts (self->ppd));
+  return PyLong_FromLong (ppdConflicts (self->ppd));
 }
 
 static PyObject *
 PPD_findOption (PPD *self, PyObject *args)
 {
   PyObject *ret;
-  const char *option;
+  PyObject *optionobj;
+  char *option;
   ppd_option_t *opt;
 
-  if (!PyArg_ParseTuple (args, "s", &option))
+  if (!PyArg_ParseTuple (args, "O", &optionobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&option, optionobj) == NULL)
     return NULL;
 
   opt = ppdFindOption (self->ppd, option);
+  free (option);
   if (opt) {
     PyObject *args = Py_BuildValue ("()");
     PyObject *kwlist = Py_BuildValue ("{}");
@@ -325,8 +409,7 @@ PPD_findOption (PPD *self, PyObject *args)
     Py_INCREF (self);
     ret = (PyObject *) optobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -336,16 +419,30 @@ static PyObject *
 PPD_findAttr (PPD *self, PyObject *args, PyObject *kwds)
 {
   PyObject *ret;
-  const char *name;
-  const char *spec = NULL;
+  PyObject *nameobj;
+  PyObject *specobj = NULL;
+  char *name;
+  char *spec = NULL;
   ppd_attr_t *attr;
   static char *kwlist[] = { "name", "spec", NULL };
 
-  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
-				    &name, &spec))
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &nameobj, &specobj))
     return NULL;
 
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
+    return NULL;
+
+  if (specobj)
+    if (UTF8_from_PyObj (&spec, specobj) == NULL) {
+      free (name);
+      return NULL;
+    }
+
   attr = ppdFindAttr (self->ppd, name, spec);
+  free (name);
+  if (spec)
+    free (spec);
   if (attr) {
     PyObject *largs = Py_BuildValue ("()");
     PyObject *lkwlist = Py_BuildValue ("{}");
@@ -358,8 +455,7 @@ PPD_findAttr (PPD *self, PyObject *args, PyObject *kwds)
     Py_INCREF (self);
     ret = (PyObject *) attrobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -369,16 +465,30 @@ static PyObject *
 PPD_findNextAttr (PPD *self, PyObject *args, PyObject *kwds)
 {
   PyObject *ret;
-  const char *name;
-  const char *spec = NULL;
+  PyObject *nameobj;
+  PyObject *specobj = NULL;
+  char *name;
+  char *spec = NULL;
   ppd_attr_t *attr;
   static char *kwlist[] = { "name", "spec", NULL };
 
-  if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
-				    &name, &spec))
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "O|O", kwlist,
+				    &nameobj, &specobj))
     return NULL;
 
+  if (UTF8_from_PyObj (&name, nameobj) == NULL)
+    return NULL;
+
+  if (specobj)
+    if (UTF8_from_PyObj (&spec, specobj) == NULL) {
+      free (name);
+      return NULL;
+    }
+
   attr = ppdFindNextAttr (self->ppd, name, spec);
+  free (name);
+  if (spec)
+    free (spec);
   if (attr) {
     PyObject *largs = Py_BuildValue ("()");
     PyObject *lkwlist = Py_BuildValue ("{}");
@@ -391,8 +501,7 @@ PPD_findNextAttr (PPD *self, PyObject *args, PyObject *kwds)
     Py_INCREF (self);
     ret = (PyObject *) attrobj;
   } else {
-    ret = Py_None;
-    Py_INCREF (ret);
+    Py_RETURN_NONE;
   }
 
   return ret;
@@ -451,6 +560,184 @@ PPD_nondefaultsMarked (PPD *self)
   return PyBool_FromLong (nondefaults_marked);
 }
 
+#if !(_POSIX_C_SOURCE >= 200809L || _XOPEN_SOURCE >= 700)
+/*
+ * A rudimentary emulation of getline() for systems that dont support it
+ * natively.  Since this is used for PPD file reading, it assumes (possibly
+ * falsely) that BUFSIZ is big enough.
+ */
+ssize_t
+getline(char **line, size_t *linelen, FILE *fp)
+{
+  if (*linelen == 0) {
+    *linelen = BUFSIZ;
+    *line = malloc(*linelen);
+  }
+
+  memset(*line, 0, *linelen);
+  fgets(*line, *linelen, fp);
+
+  return (strlen(*line));
+}
+#endif
+
+/*
+ * emit marked options by returning a string.
+ */
+
+static PyObject *
+PPD_emitString (PPD *self, PyObject *args)
+{
+  ppd_section_t section;
+  float min_order;
+  char *emitted;
+  PyObject *ret;
+
+  if (!PyArg_ParseTuple (args, "if", &section, &min_order))
+    return NULL;
+
+  emitted = ppdEmitString(self->ppd, section, min_order);
+
+  if (emitted) {
+    ret = PyUnicode_FromString (emitted);
+    free (emitted);
+  } else {
+    Py_RETURN_NONE;
+  }
+
+  return ret;
+}
+
+/*
+ * emit marked options by writing to a file object.
+ */
+
+static PyObject *
+PPD_emit (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  ppd_section_t section;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "Oi", &pyFile, &section))
+    return NULL;
+
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+  if (!f)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
+
+  if (!ppdEmit(self->ppd, f, section))
+    Py_RETURN_NONE;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit marked options after order dependency by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitAfterOrder (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  ppd_section_t section;
+  FILE *f;
+  int limit;
+  float min_order;
+
+  if (!PyArg_ParseTuple (args, "Oiif", &pyFile, &section, &limit, &min_order))
+    return NULL;
+
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+  if (!f)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
+
+  if (!ppdEmitAfterOrder(self->ppd, f, section, limit, min_order))
+    Py_RETURN_NONE;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit marked options by writing to a file descriptor.
+ */
+
+static PyObject *
+PPD_emitFd (PPD *self, PyObject *args)
+{
+  ppd_section_t section;
+  int f;
+
+  if (!PyArg_ParseTuple (args, "ii", &f, &section))
+    return NULL;
+
+  if (!ppdEmitFd(self->ppd, f, section))
+    Py_RETURN_NONE;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit JCL options by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitJCL (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  int job_id;
+  PyObject *userobj;
+  PyObject *titleobj;
+  char *user;
+  char *title;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "OiOO", &pyFile, &job_id, &userobj, &titleobj))
+    return NULL;
+
+  if (UTF8_from_PyObj (&user, userobj) == NULL)
+    return NULL;
+
+  if (UTF8_from_PyObj (&title, titleobj) == NULL) {
+    free (user);
+    return NULL;
+  }
+
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+  if (!f)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
+
+  if (!ppdEmitJCL(self->ppd, f, job_id, user, title))
+    Py_RETURN_NONE;
+
+  free (user);
+  free (title);
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
+/*
+ * emit JCL end by writing to a file object.
+ */
+
+static PyObject *
+PPD_emitJCLEnd (PPD *self, PyObject *args)
+{
+  PyObject *pyFile;
+  FILE *f;
+
+  if (!PyArg_ParseTuple (args, "O", &pyFile))
+    return NULL;
+
+  int fd = PyObject_AsFileDescriptor(pyFile);
+  f = fdopen(fd, "w");
+  if (!f)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
+
+  if (!ppdEmitJCLEnd(self->ppd, f))
+    Py_RETURN_NONE;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
 PyObject *
 PPD_writeFd (PPD *self, PyObject *args)
 {
@@ -463,16 +750,12 @@ PPD_writeFd (PPD *self, PyObject *args)
     return NULL;
 
   dfd = dup (fd);
-  if (!dfd) {
-    PyErr_SetFromErrno (PyExc_RuntimeError);
-    return NULL;
-  }
+  if (dfd == -1)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   out = fdopen (dfd, "w");
-  if (!out) {
-    PyErr_SetFromErrno (PyExc_RuntimeError);
-    return NULL;
-  }
+  if (!out)
+    return PyErr_SetFromErrno (PyExc_RuntimeError);
 
   rewind (self->file);
   while (!feof (self->file)) {
@@ -489,7 +772,8 @@ PPD_writeFd (PPD *self, PyObject *args)
       for (end = start; *end; end++)
 	if (isspace (*end) || *end == ':')
 	  break;
-      keyword = strndup (start, end-start);
+      keyword = calloc(1, (end - start) + 1);
+      strncpy(keyword, start, end - start);
       choice = ppdFindMarkedChoice (self->ppd, keyword);
 
       // Treat PageRegion, PaperDimension and ImageableArea specially:
@@ -516,8 +800,7 @@ PPD_writeFd (PPD *self, PyObject *args)
   if (line)
     free (line);
 
-  Py_INCREF (Py_None);
-  return Py_None;
+  Py_RETURN_NONE;
 }
 
 /////////
@@ -619,6 +902,16 @@ PyMethodDef PPD_methods[] =
       "localize() -> None\n\n"
       "Localize PPD to the current locale." },
 
+    { "localizeIPPReason",
+      (PyCFunction) PPD_localizeIPPReason, METH_VARARGS | METH_KEYWORDS,
+      "localizeIPPReason(reason, scheme) -> string or None\n\n"
+      "Localize IPP reason to the current locale." },
+
+    { "localizeMarkerName",
+      (PyCFunction) PPD_localizeMarkerName, METH_VARARGS,
+      "localizeMarkerName(name) -> string or None\n\n"
+      "Localize marker name to the current locale." },
+
     { "markDefaults",
       (PyCFunction) PPD_markDefaults, METH_NOARGS,
       "markDefaults() -> None\n\n"
@@ -672,6 +965,67 @@ PyMethodDef PPD_methods[] =
       "nondefaultsMarked() -> boolean\n\n"
       "Returns true if any non-default option choices are marked." },
 
+    { "emitString",
+      (PyCFunction) PPD_emitString, METH_VARARGS,
+      "emitString(section, min_order) -> string\n\n"
+      "Return a string with the marked options for section, with at least min_order order dependency.\n"
+      "@type section: integer\n"
+      "@param section: section id\n"
+      "@type min_order: float\n"
+      "@param min_order: minimum order dependency\n"
+      "@return: string containing emitted postscript" },
+
+    { "emit",
+      (PyCFunction) PPD_emit, METH_VARARGS,
+      "emit(file, section) -> None\n\n"
+      "Output marked options for section to a file.\n"
+      "@type file: file\n"
+      "@param file: file object\n"
+      "@type section: integer\n"
+      "@param section: section id" },
+
+    { "emitAfterOrder",
+      (PyCFunction) PPD_emitAfterOrder, METH_VARARGS,
+      "emitAfterOrder(file, section, limit, min_order) -> None\n\n"
+      "Output marked options for section to a file.\n"
+      "@type file: file\n"
+      "@param file: file object\n"
+      "@type section: integer\n"
+      "@param section: section id\n"
+      "@type limit: integer\n"
+      "@param limit: non-zero to use min_order\n"
+      "@type min_order: float\n"
+      "@param min_order: minimum order dependency" },
+
+    { "emitFd",
+      (PyCFunction) PPD_emitFd, METH_VARARGS,
+      "emitFd(fd, section) -> None\n\n"
+      "Output marked options for section to a file descriptor.\n"
+      "@type fd: integer\n"
+      "@param fd: file descriptor\n"
+      "@type section: integer\n"
+      "@param section: section id" },
+
+    { "emitJCL",
+      (PyCFunction) PPD_emitJCL, METH_VARARGS,
+      "emitJCL(file, job_id, user, title) -> None\n\n"
+      "Emit code for JCL options to a file.\n"
+      "@type file: file object\n"
+      "@param file: file\n"
+      "@type job_id: integer\n"
+      "@param job_id: job id\n"
+      "@type user: string\n"
+      "@param user: user name on job\n"
+      "@type title: string\n"
+      "@param title: title of job" },
+
+    { "emitJCLEnd",
+      (PyCFunction) PPD_emitJCLEnd, METH_VARARGS,
+      "emitJCLEnd(file) -> None\n\n"
+      "Emit JCLEnd code to a file.\n"
+      "@type file: file object\n"
+      "@param file: file" },
+
     { "writeFd",
       (PyCFunction) PPD_writeFd, METH_VARARGS,
       "writeFd(fd) -> None\n\n"
@@ -685,8 +1039,7 @@ PyMethodDef PPD_methods[] =
 
 PyTypeObject cups_PPDType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.PPD",                /*tp_name*/
     sizeof(PPD),               /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -758,7 +1111,20 @@ static void
 Option_dealloc (Option *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
+}
+
+static PyObject *
+Option_repr (Option *self)
+{
+  ppd_option_t *option = self->option;
+  if (!option)
+    return PyUnicode_FromString ("<cups.Option>");
+
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Option %s=%s>",
+			  option->keyword, option->defchoice);
+    return PyUnicode_FromString (buffer);
 }
 
 ////////////
@@ -778,8 +1144,7 @@ static PyObject *
 Option_getKeyword (Option *self, void *closure)
 {
   if (!self->option) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->option->keyword);
@@ -789,8 +1154,7 @@ static PyObject *
 Option_getDefchoice (Option *self, void *closure)
 {
   if (!self->option) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->option->defchoice);
@@ -800,8 +1164,7 @@ static PyObject *
 Option_getText (Option *self, void *closure)
 {
   if (!self->option) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->option->text);
@@ -811,11 +1174,10 @@ static PyObject *
 Option_getUI (Option *self, void *closure)
 {
   if (!self->option) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
-  return PyInt_FromLong (self->option->ui);
+  return PyLong_FromLong (self->option->ui);
 }
 
 static PyObject *
@@ -909,8 +1271,7 @@ PyMethodDef Option_methods[] =
 
 PyTypeObject cups_OptionType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Option",             /*tp_name*/
     sizeof(Option),            /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -919,7 +1280,7 @@ PyTypeObject cups_OptionType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Option_repr,     /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
@@ -989,7 +1350,19 @@ static void
 Group_dealloc (Group *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
+}
+
+static PyObject *
+Group_repr (Group *self)
+{
+  ppd_group_t *group = self->group;
+  if (!group)
+    return PyUnicode_FromString ("<cups.Group>");
+
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Group %s>", group->name);
+    return PyUnicode_FromString (buffer);
 }
 
 ///////////
@@ -1000,8 +1373,7 @@ static PyObject *
 Group_getText (Group *self, void *closure)
 {
   if (!self->group) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->group->text);
@@ -1011,8 +1383,7 @@ static PyObject *
 Group_getName (Group *self, void *closure)
 {
   if (!self->group) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->group->name);
@@ -1102,8 +1473,7 @@ PyMethodDef Group_methods[] =
 
 PyTypeObject cups_GroupType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Group",              /*tp_name*/
     sizeof(Group),             /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1112,7 +1482,7 @@ PyTypeObject cups_GroupType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Group_repr,      /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
@@ -1177,7 +1547,7 @@ static void
 Constraint_dealloc (Constraint *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
 }
 
 ////////////////
@@ -1188,8 +1558,7 @@ static PyObject *
 Constraint_getOption1 (Constraint *self, void *closure)
 {
   if (!self->constraint) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->constraint->option1);
@@ -1199,8 +1568,7 @@ static PyObject *
 Constraint_getChoice1 (Constraint *self, void *closure)
 {
   if (!self->constraint) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->constraint->choice1);
@@ -1210,8 +1578,7 @@ static PyObject *
 Constraint_getOption2 (Constraint *self, void *closure)
 {
   if (!self->constraint) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->constraint->option2);
@@ -1221,8 +1588,7 @@ static PyObject *
 Constraint_getChoice2 (Constraint *self, void *closure)
 {
   if (!self->constraint) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->constraint->choice2);
@@ -1251,8 +1617,7 @@ PyGetSetDef Constraint_getseters[] =
 
 PyTypeObject cups_ConstraintType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Constraint",         /*tp_name*/
     sizeof(Constraint),        /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1326,7 +1691,22 @@ static void
 Attribute_dealloc (Attribute *self)
 {
   Py_XDECREF (self->ppd);
-  self->ob_type->tp_free ((PyObject *) self);
+  ((PyObject *)self)->ob_type->tp_free ((PyObject *) self);
+}
+
+static PyObject *
+Attribute_repr (Attribute *self)
+{
+  ppd_attr_t *attribute = self->attribute;
+  if (!attribute)
+    return PyUnicode_FromString ("<cups.Attribute>");
+
+  char buffer[256];
+  snprintf (buffer, 256, "<cups.Attribute *%s%s%s>",
+			  attribute->name,
+			  attribute->spec[0] != '\0' ? " ": "",
+			  attribute->spec);
+    return PyUnicode_FromString (buffer);
 }
 
 ///////////////
@@ -1337,8 +1717,7 @@ static PyObject *
 Attribute_getName (Attribute *self, void *closure)
 {
   if (!self->attribute) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->attribute->name);
@@ -1348,8 +1727,7 @@ static PyObject *
 Attribute_getSpec (Attribute *self, void *closure)
 {
   if (!self->attribute) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->attribute->spec);
@@ -1359,8 +1737,7 @@ static PyObject *
 Attribute_getText (Attribute *self, void *closure)
 {
   if (!self->attribute) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->attribute->text);
@@ -1370,8 +1747,7 @@ static PyObject *
 Attribute_getValue (Attribute *self, void *closure)
 {
   if (!self->attribute) {
-    Py_INCREF (Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
   }
 
   return make_PyUnicode_from_ppd_string (self->ppd, self->attribute->value);
@@ -1400,8 +1776,7 @@ PyGetSetDef Attribute_getseters[] =
 
 PyTypeObject cups_AttributeType =
   {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "cups.Attribute",          /*tp_name*/
     sizeof(Attribute),         /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1410,7 +1785,7 @@ PyTypeObject cups_AttributeType =
     0,                         /*tp_getattr*/
     0,                         /*tp_setattr*/
     0,                         /*tp_compare*/
-    0,                         /*tp_repr*/
+    (reprfunc)Attribute_repr,  /*tp_repr*/
     0,                         /*tp_as_number*/
     0,                         /*tp_as_sequence*/
     0,                         /*tp_as_mapping*/
